use anyhow::Result;
use num_bigint::BigInt;
use num_traits::Num;

use crate::weierstrass_curve::{Point, WeierstrassCurve};

pub struct EllipticCurve {
    pub curve: WeierstrassCurve,
    pub gen_point: Point,
    pub pub_point: Point,
}

pub struct PrivateKey {
    pub gen_order: BigInt,
    pub private_key: BigInt,
}

impl PrivateKey {
    pub fn new(gen_order: &str, private_key: &str) -> Result<Self> {
        let gen_order = BigInt::from_str_radix(gen_order, 10)?;
        let private_key = BigInt::from_str_radix(private_key, 10)?;
        Ok(Self {
            gen_order,
            private_key,
        })
    }
}

impl EllipticCurve {
    pub fn new(
        p: &str,
        a: &str,
        generator_x: &str,
        generator_y: &str,
        public_key_x: &str,
        public_key_y: &str,
    ) -> Result<Self> {
        let p = BigInt::from_str_radix(p, 10)?;
        let a = BigInt::from_str_radix(a, 10)?;
        let generator_x = BigInt::from_str_radix(generator_x, 10)?;
        let generator_y = BigInt::from_str_radix(generator_y, 10)?;
        let public_key_x = BigInt::from_str_radix(public_key_x, 10)?;
        let public_key_y = BigInt::from_str_radix(public_key_y, 10)?;

        let curve = WeierstrassCurve::new(a, p);

        let gen_point = Point::Point {
            x: generator_x,
            y: generator_y,
        };

        let pub_point = Point::Point {
            x: public_key_x,
            y: public_key_y,
        };

        Ok(Self {
            curve,
            gen_point,
            pub_point,
        })
    }
}
